#include <gccore.h>
#include <stdio.h>
#include <malloc.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <fat.h>

#include <wiiuse/wpad.h>
#include <gba.h>	// Header for GBA functions

static void *xfb = NULL;
static GXRModeObj *rmode = NULL;

typedef struct _txtNames {
	char name[256];
} txtNames;

int compare (const void * a, const void * b ) {
  return strcmp((*(txtNames*)a).name, (*(txtNames*)b).name);
}

//---------------------------------------------------------------------------------
int main(int argc, char **argv) {
//---------------------------------------------------------------------------------
	int i = 0, j = 0, k = 0;
	int txtCnt = 0;
	bool initGBA = false;

	// Initialise the fat devices
	fatInitDefault();

	// Initialise the video system
	VIDEO_Init();
	
	// This function initialises the attached controllers
	PAD_Init();
	WPAD_Init();
	
	// Initialise the GameBoy Advance
	//intitGBA = GBA_Init();
	if(GBA_Init())
		initGBA = true;

	// Enable controller mode with screen off
	GBA_ScanPads(LCD_OFF);
	
	// Obtain the preferred video mode from the system
	// This will correspond to the settings in the Wii menu
	rmode = VIDEO_GetPreferredMode(NULL);

	// Allocate memory for the display in the uncached region
	xfb = MEM_K0_TO_K1(SYS_AllocateFramebuffer(rmode));
	
	// Initialise the console, required for printf
	console_init(xfb,20,20,rmode->fbWidth,rmode->xfbHeight,rmode->fbWidth*VI_DISPLAY_PIX_SZ);
	
	// Set up the video registers with the chosen mode
	VIDEO_Configure(rmode);
	
	// Tell the video hardware where our display memory is
	VIDEO_SetNextFramebuffer(xfb);
	
	// Make the display visible
	VIDEO_SetBlack(FALSE);

	// Flush the video register changes to the hardware
	VIDEO_Flush();

	// Wait for Video setup to complete
	VIDEO_WaitVSync();
	if(rmode->viTVMode&VI_NON_INTERLACE) VIDEO_WaitVSync();

	// Read and sort test directory based on .txt extension
	DIR *dir = opendir("/test");
	struct dirent *dent;
	txtNames *names;

	if(dir!=NULL)
	{
		while((dent=readdir(dir))!=NULL)
		{
			if(strstr(dent->d_name,".txt") != NULL)
				txtCnt++;
		}
		closedir(dir);
		names = malloc(sizeof(txtNames)*txtCnt);
		memset(names,0,sizeof(txtNames)*txtCnt);
		dir = opendir("/test");
		int i = 0;
		while((dent=readdir(dir))!=NULL)
		{
			if(strstr(dent->d_name,".txt") != NULL)
			{
				strcpy(names[i].name,dent->d_name);
				i++;
			}
		}
		closedir(dir);
	}

	if(txtCnt == 0)
	{
		printf("No Files! Make sure you have .txt files in your \"test\" folder!\n");
		VIDEO_WaitVSync();
		VIDEO_WaitVSync();
		sleep(1);
		return 0;
	}
	qsort(names, txtCnt, sizeof(txtNames), compare);

	if(!initGBA)
		printf("\nGameBoy Advance not found!\n");

	while(1) {
		int done = 0;
		int send = 0;
		while(!done)
		{
			printf("\x1b[2J");
			printf("\x1b[3;1HSelect a .txt file and press A :");
			printf("\n<< %s >>\n",names[i].name);

			printf("\x1b[22;0HGC Z / Wiimote + : Reset GBA.(Text mode only)\n");
			printf("GC L / Wiimote - : Clear GBA screen.\n");
			printf("GC X / Wiimote 1 : GBA controller mode.\n");
			printf("GC Y / Wiimote 2 : GBA text mode.\n");
			printf("GC Start / Wiimote Home : Exit.\n");

			WPAD_ScanPads();
			PAD_ScanPads();
			VIDEO_WaitVSync();

			u32 btnsWii = WPAD_ButtonsDown(0);
			u32 btnsGC = PAD_ButtonsDown(0);

			if(btnsGC & PAD_BUTTON_A || btnsWii & WPAD_BUTTON_A)
			{
				done = 1;
				send = 1;
				break;
			}
			else if(btnsGC & PAD_BUTTON_RIGHT || btnsWii & WPAD_BUTTON_RIGHT)
			{
				i++;
				if(i >= txtCnt) i = 0;
			}
			else if(btnsGC & PAD_BUTTON_LEFT || btnsWii & WPAD_BUTTON_LEFT)
			{
				i--;
				if(i < 0) i = (txtCnt-1);
			}	
			else if(btnsGC & PAD_BUTTON_START || PAD_ButtonsHeld(1) & PAD_BUTTON_START || btnsWii & WPAD_BUTTON_HOME)
			{
				VIDEO_WaitVSync();
				VIDEO_WaitVSync();
				exit(0);
			}
			else if(btnsGC & PAD_TRIGGER_Z || btnsWii & WPAD_BUTTON_PLUS)
			{
				/* Reset GBA on GameCube port 0 to 3 */
				printf("\x1b[6;0HRestarting the GameBoy Advance...\n");
				VIDEO_WaitVSync();

				for(j=0; j<4; j++)
					GBA_Reset(j);
				GBA_Init();
				done = 1;
			}
			else if(btnsGC & PAD_TRIGGER_L || btnsWii & WPAD_BUTTON_MINUS)
			{
				/* Clear GBA screen on port 0 to 1 */
				for(j=0; j<2; j++)
					GBA_ClearScreen(j);
				done = 1;
			}
			else if(btnsGC & PAD_BUTTON_X || btnsWii & WPAD_BUTTON_1)
			{
				/* Enable GBA controller mode */
				for(j=0; j<2; j++)
					GBA_ScanPads(LCD_OFF);
				done = 1;
			}
			else if(btnsGC & PAD_BUTTON_Y || btnsWii & WPAD_BUTTON_2)
			{
				/* Enable GBA text mode */
				for(j=0; j<2; j++)
					GBA_TextMode(j);
				done = 1;
			}
		}
		/* File selected : send text to the GameBoy Advance */
		if(send)
		{
			char textfile[256];
			sprintf(textfile,"/test/%s",names[i].name);

			/* Enable text mode */
			GBA_TextMode(0);

			printf("\x1b[6;0HSending text file to the GameBoy Advance...\n");
			VIDEO_WaitVSync();

			/* Print text on GameCube port 0 */
			GBA_PrintText("Receiving text file.\n", 0);

			char msg[256]="\nLoading ...";

			/* Print a message character by character on GameCube port 0 */
			for(k=0; k<strlen(msg); k++)
				GBA_PrintChar(msg[k], 0, 50);

			/* Print selected file on GameCube port 0 */
			GBA_PrintFile(textfile, 0);
		}

		// Wait for the next frame
		VIDEO_WaitVSync();
	}

	return 0;
}
